#!/usr/bin/env python3

import logging as logging
from logging import handlers
import os


class Logger(logging.Logger):

    def __init__(self, name, logdir, envtype=None):
        super().__init__(name=name, level=logging.DEBUG)

        # create console handler and set level to info
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        # create file handler and set level to debug
        logfile = os.path.join(logdir, name + '.log')
        fh = handlers.TimedRotatingFileHandler(filename=logfile, when='D', interval=1, backupCount=6)
        fh.setLevel(logging.DEBUG)
        # create smtp handler and set level to error
        sh = handlers.SMTPHandler(mailhost='smtp.inria.fr', fromaddr='logger@gtlville.inrialpes.fr',
                                  toaddrs=['dance.eng@inria.fr'], subject='GTLVille alert')
        sh.setLevel(logging.ERROR)

        # create formatter
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        # add formatter to handlers
        ch.setFormatter(formatter)
        fh.setFormatter(formatter)
        sh.setFormatter(formatter)

        # add handlers to logger
        self.addHandler(ch)
        self.addHandler(fh)
        if envtype != 'local':
            self.addHandler(sh)



#!/usr/bin/env python3

import asyncio
import os
import traceback
from typing import Any, Awaitable, Callable

from gv_pubsub.pubsub import PubSub
from gv_requests.requests import Requests
from gv_utils.asyncio import check_event_loop
from gv_utils.datetime import datetime
from gv_utils.enums import MessageKind
from gv_utils.geometry import BaseGeometry
from gv_utils.logger import Logger


class Service:
    samplings = {MessageKind.karrusrd: 1 * 60, MessageKind.metropme: 1 * 60, MessageKind.tomtomfcd: 1 * 60}

    def __init__(self, logger: Logger, futures: list = None, callbacks: dict = None,
                 custom_data_decoder: Callable = None, offlineprefix: str = None) -> None:
        if futures is None:
            futures = []
        if callbacks is None:
            callbacks = {}
        if offlineprefix is None:
            offlineprefix = 'OFFLINE'

        self.logger = logger
        self.futures = futures
        self.callbacks = callbacks
        self.pubsub = PubSub(self.logger, custom_data_decoder)
        self.requests = Requests(logger)
        self.offlineprefix = offlineprefix
        self._mainfut = None

    async def async_init(self) -> None:
        pass

    def start(self, redisaddr: str, apiaddr: str) -> None:
        check_event_loop()  # will create a new event loop if needed (if we are not in the main thread)
        self.logger.info('Service is starting...')
        try:
            asyncio.run(self._run(redisaddr, apiaddr))
        except KeyboardInterrupt:
            pass
        self.logger.info('Service has stopped.')

    async def _run(self, redisaddr: str, apiaddr: str) -> None:
        try:
            await asyncio.gather(self.pubsub.async_init(redisaddr), self.requests.async_init(apiaddr))
            await self.async_init()
            self.logger.info('Service has started.')
            try:
                self._mainfut = asyncio.gather(
                    *self.futures,
                    *[self._subscribe(datakind, callback) for datakind, callback in self.callbacks.items()]
                )
                await self._mainfut
            except (KeyboardInterrupt, asyncio.CancelledError):
                self._cancel()
            except:
                self.logger.error('An error occurred in the main task. Exiting!{}{}'.format(os.linesep,
                                                                                            traceback.format_exc()))
            finally:
                await self._close()
        except:
            self.logger.error('An error occurred at init. Exiting!{}{}'.format(os.linesep,
                                                                               traceback.format_exc()))
            await self._close()

    async def _close(self) -> None:
        await asyncio.gather(self.pubsub.close(), self.requests.close())

    def _cancel(self) -> None:
        if self._mainfut is not None:
            self._mainfut.cancel()
            self._mainfut = None

    async def _publish(self, data: object, dataformat: str, datatimestamp, datakind: str) -> bool:
        return await self.pubsub.publish(data, dataformat, datatimestamp, datakind)

    async def _subscribe(self, datakind: str, callback: Callable[[dict], Awaitable]) -> None:
        await self.pubsub.subscribe(datakind, callback)

    async def _get_data_point(self, datapointeids: list = None, datatypeeids: list = None,
                              area: BaseGeometry = None) -> list:
        return await self.requests.get_data_point(datapointeids, datatypeeids, area)

    async def _get_road(self, roadeids: list = None, area: BaseGeometry = None) -> list:
        return await self.requests.get_road(roadeids, area)

    async def _get_road_center(self, roadeids: list = None, area: BaseGeometry = None) -> list:
        return await self.requests.get_road_center(roadeids, area)

    async def _get_turn_ratio(self, fromroad: str = None, toroad: str = None) -> list:
        return await self.requests.get_turn_ratio(fromroad, toroad)

    async def _get_prediction_point(self) -> list:
        return await self.requests.get_prediction_point()

    async def _get_zone_point(self, zoneeids: list = None, area: BaseGeometry = None) -> list:
        return await self.requests.get_zone_point(zoneeids, area)

    async def _get_road_data_point(self, datatype: str = None, area: BaseGeometry = None) -> list:
        return await self.requests.get_road_data_point(datatype, area)

    async def _get_data_point_data_quality(self, fromdatetime: datetime, todatetime: datetime = None,
                                           datapointeids: list = None, datatypeeids: list = None, period: str = None,
                                           sampling: str = None, window: str = None) -> dict:
        return await self.requests.get_data_point_data_quality(fromdatetime, todatetime, datapointeids, datatypeeids,
                                                               period, sampling, window)

    async def _get_data_type_data_quality(self, fromdatetime: datetime, todatetime: datetime = None,
                                          datatypeeids: list = None, period: str = None, sampling: str = None,
                                          window: str = None) -> dict:
        return await self.requests.get_data_type_data_quality(fromdatetime, todatetime, datatypeeids, period, sampling,
                                                              window)

    async def _get_road_data_quality(self, fromdatetime: datetime, todatetime: datetime = None, roadeids: list = None,
                                     period: str = None, sampling: str = None, window: str = None) -> dict:
        return await self.requests.get_road_data_quality(fromdatetime, todatetime, roadeids, period, sampling, window)

    async def _get_data_point_indicator(self, indicator: str, fromdatetime: datetime, todatetime: datetime = None,
                                        datapointeids: list = None, datatypeeids: list = None, period: str = None,
                                        sampling: str = None, window: str = None) -> dict:
        return await self.requests.get_data_point_indicator(indicator, fromdatetime, todatetime, datapointeids,
                                                            datatypeeids, period, sampling, window)

    async def _get_road_indicator(self, indicator: str, fromdatetime: datetime, todatetime: datetime = None,
                                  roadeids: list = None, period: str = None, sampling: str = None,
                                  window: str = None) -> dict:
        return await self.requests.get_road_indicator(indicator, fromdatetime, todatetime, roadeids, period, sampling,
                                                      window)

    async def _get_zone_travel_time(self, fromdatetime: datetime, todatetime: datetime = None, period: str = None,
                                    sampling: str = None, frompointeid: str = None, topointeid: str = None) -> dict:
        return await self.requests.get_zone_travel_time(fromdatetime, todatetime, period,
                                                        sampling, frompointeid, topointeid)

    @staticmethod
    async def _get_not_empty(get_func: Callable, args: tuple = (), kwargs: dict = None) -> Any:
        if kwargs is None:
            kwargs = {}

        res = {}
        while len(res) == 0:
            res = await get_func(*args, **kwargs)
        return res


def start(Application, threaded=False):
    if threaded:
        import threading
        threading.Thread(target=start, args=(Application, False), daemon=True).start()
        print('Starting application in a background thread...')
    else:
        Application()



#!/usr/bin/env python3


# see https://stackoverflow.com/questions/34988692/python-3-multiprocessing-optimal-chunk-size
def get_chunksize(size: int, ncpu: int, taskpt: int, chunkpt: int = 10000) -> int:  # taskpt & chunkpt in microseconds
    return max(1, min(_div(chunkpt, taskpt), _div(size, ncpu)))


def _div(x: int, y: int) -> int:
    return x // y + min(1, x % y)



#!/usr/bin/env python3

import math
from typing import Generator, Iterable, List, Tuple, Union

import shapely.geometry
from shapely.geometry.base import BaseGeometry
import shapely.ops
import shapely.wkb
import shapely.wkt

LONLAT_CENTER = (5.74626, 45.17475)
LONLAT_EXTEND = ((5.60268, 45.0821), (5.88934, 45.2674))
EXTEND_BBOX_XY = None
SRID = 4326
R = 6371000  # earth radius in meters
ONE_METER_IN_LATLON = .0000128


def geom_to_xy(geom: BaseGeometry, lonlatcenter: Tuple[float, float] = None, rounding: int = 2) -> BaseGeometry:
    if lonlatcenter is None:
        lonlatcenter = LONLAT_CENTER

    return shapely.ops.transform(lambda lon, lat, z=None: _to_xy(lon, lat, lonlatcenter, rounding), geom)


def lonlat_to_xy(lonlat: Tuple[float, float], lonlatcenter: Tuple[float, float] = None,
                 rounding: int = 2) -> Tuple[float, float]:
    if lonlatcenter is None:
        lonlatcenter = LONLAT_CENTER

    return _to_xy(*lonlat, lonlatcenter=lonlatcenter, rounding=rounding)


def coords_to_xy(coords: Iterable[Tuple[float, float]], lonlatcenter: Tuple[float, float] = None,
                 rounding: int = 2) -> Generator[Tuple[float, float], None, None]:
    if lonlatcenter is None:
        lonlatcenter = LONLAT_CENTER

    for lonlat in coords:
        yield _to_xy(*lonlat, lonlatcenter=lonlatcenter, rounding=rounding)


def _to_xy(lon: float, lat: float, lonlatcenter: Tuple[float, float], rounding: int) -> Tuple[float, float]:
    if rounding is not None:
        _f = lambda u: round(u, rounding)
    else:
        _f = lambda u: u

    lon = lon * math.pi / 180
    lat = lat * math.pi / 180
    loncenter = lonlatcenter[0] * math.pi / 180
    latcenter = lonlatcenter[1] * math.pi / 180

    x = R * math.cos(lat) * math.sin(lon - loncenter)
    y = R * (math.cos(latcenter) * math.sin(lat) - math.sin(latcenter) * math.cos(lat) * math.cos(lon - loncenter))
    return _f(x), _f(y)


def geom_to_lonlat(geom: BaseGeometry, lonlatcenter: Tuple[float, float] = None, rounding: int = 5) -> BaseGeometry:
    if lonlatcenter is None:
        lonlatcenter = LONLAT_CENTER

    return shapely.ops.transform(lambda x, y, z=None: _to_lonlat(x, y, lonlatcenter, rounding), geom)


def xy_to_lonlat(x: float, y: float, lonlatcenter: Tuple[float, float] = None,
                 rounding: int = 5) -> Tuple[float, float]:
    if lonlatcenter is None:
        lonlatcenter = LONLAT_CENTER

    return _to_lonlat(x, y, lonlatcenter, rounding)


def coords_to_lonlat(coords: List[Tuple[float, float]], lonlatcenter: Tuple[float, float] = None,
                     rounding: int = 5) -> Generator[Tuple[float, float], None, None]:
    if lonlatcenter is None:
        lonlatcenter = LONLAT_CENTER

    for xy in coords:
        yield _to_lonlat(*xy, lonlatcenter=lonlatcenter, rounding=rounding)


def _to_lonlat(x: float, y: float, lonlatcenter: Tuple[float, float], rounding: int) -> Tuple[float, float]:
    if rounding is not None:
        _f = lambda u: round(u, rounding)
    else:
        _f = lambda u: u

    loncenter = lonlatcenter[0] * math.pi / 180
    latcenter = lonlatcenter[1] * math.pi / 180

    p = math.pow(x ** 2 + y ** 2, .5)
    c = math.asin(p / R)

    lon = loncenter + math.atan2(x * math.sin(c),
                                 p * math.cos(c) * math.cos(latcenter) - y * math.sin(c) * math.sin(latcenter))
    lat = math.asin(math.cos(c) * math.sin(latcenter) + y * math.sin(c) * math.cos(latcenter) / p)
    lon = lon / math.pi * 180
    lat = lat / math.pi * 180
    return _f(lon), _f(lat)


def round_lonlat(lonlat: Tuple[float, float]) -> Tuple[int, int]:
    return round(lonlat[0] * 1000), round(lonlat[1] * 1000)


def shift_to_right(geom: shapely.geometry.LineString, dist: int) -> shapely.geometry.LineString:
    geom = geom.parallel_offset(dist, 'right')
    geom.coords = list(geom.coords)[::-1]
    return geom


def is_inside_extend(geomxy: BaseGeometry) -> bool:
    global EXTEND_BBOX_XY
    if EXTEND_BBOX_XY is None:
        coords = []
        for coord in coords_to_xy(LONLAT_EXTEND):
            coords.extend(coord)
        EXTEND_BBOX_XY = shapely.geometry.box(*coords)

    return EXTEND_BBOX_XY.intersects(geomxy)


def close_elems_it(lonlatrounded: Tuple[int, int], lonlatmatrix: dict,
                   radius: int = 1) -> Generator[BaseGeometry, None, None]:
    if radius >= 1:
        lat, lon = lonlatrounded
        for i in _get_range(lat, radius):
            if i in lonlatmatrix:
                for j in _get_range(lon, radius):
                    if j in lonlatmatrix[i]:
                        for elem in lonlatmatrix[i][j]:
                            yield elem


def _get_range(i: int, radius: int) -> List[int]:
    return sorted(range(i - radius, i + radius + 1), key=lambda x: abs(x - i))


def encode_geometry(geometry: Union[BaseGeometry, str]) -> str:
    if not isinstance(geometry, str):
        geometry = shapely.wkt.dumps(geometry)
    return '{};{}'.format('SRID={}'.format(SRID), geometry)


def encode_geometry_bytes(geometry: Union[BaseGeometry, str]) -> str:
    if not isinstance(geometry, BaseGeometry):
        geometry = decode_geometry(geometry)
    return shapely.wkb.dumps(geometry, srid=SRID)


def decode_geometry(wkt: str) -> BaseGeometry:
    geom = None
    try:
        geom = shapely.wkt.loads(wkt.split(';')[-1])
    except:
        try:
            geom = shapely.geometry.shape(wkt)
        except:
            pass
    return geom


def decode_geometry_bytes(wkb: bytes) -> BaseGeometry:
    geom = None
    try:
        geom = shapely.wkb.loads(wkb)
    except:
        pass
    return geom
